/*
************************************************************************
*
*   DefCircle.c - default circle drawing
*
*   Copyright (c) 1994
*
*   ETH Zuerich
*   Institut fuer Molekularbiologie und Biophysik
*   ETH-Hoenggerberg
*   CH-8093 Zuerich
*
*   SPECTROSPIN AG
*   Industriestr. 26
*   CH-8117 Faellanden
*
*   All Rights Reserved
*
*   Date of last modification : 96/02/27
*   Pathname of SCCS file     : /sgiext/molmol/sg/src/def/SCCS/s.DefCircle.c
*   SCCS identification       : 1.13
*
************************************************************************
*/

#include <sg_def.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <values.h>

#include <sg.h>
#include <sg_get.h>
#include <sg_map.h>
#include "def_sphere.h"

static const float EpsDist = 1.0e-6;

static int MaxPrec = -1;

typedef struct {
  int pointNo;
  float (* points2D)[2];
  float (* points0)[3];
  float (* points1)[3];
  float (* pointsR)[3];
} CircData;

static CircData *Circles;

static void
calcCirc(CircData *circP, int prec)
{
  int stepNo;
  double ang, cosAng, sinAng;
  double x0, y0, x1, y1;
  int i;

  stepNo = 1 << prec;
  circP->pointNo = 4 * stepNo + 1;

  circP->points2D = malloc(circP->pointNo * sizeof(*circP->points2D));
  circP->points0 = malloc(circP->pointNo * sizeof(*circP->points0));
  circP->points1 = malloc(circP->pointNo * sizeof(*circP->points1));
  circP->pointsR = malloc(circP->pointNo * sizeof(*circP->pointsR));

  ang = M_PI / (2 * stepNo);
  cosAng = cos(ang);
  sinAng = sin(ang);
  x1 = 1.0;
  y1 = 0.0;
  for (i = 0; i < stepNo; i++) {
    circP->points0[i][0] = x1;
    circP->points0[i][1] = y1;

    circP->points0[stepNo + i][0] = - y1;
    circP->points0[stepNo + i][1] = x1;

    circP->points0[2 * stepNo + i][0] = - x1;
    circP->points0[2 * stepNo + i][1] = - y1;

    circP->points0[3 * stepNo + i][0] = y1;
    circP->points0[3 * stepNo + i][1] = - x1;

    x0 = x1;
    y0 = y1;
    x1 = x0 * cosAng - y0 * sinAng;
    y1 = x0 * sinAng + y0 * cosAng;
  }

  /* repeat starting point at the end to draw circle as polyline */
  circP->points0[4 * stepNo][0] = circP->points0[0][0];
  circP->points0[4 * stepNo][1] = circP->points0[0][1];

  for (i = 0; i < circP->pointNo; i++) {
    circP->points0[i][2] = 0.0;

    circP->points1[i][0] = circP->points0[i][0];
    circP->points1[i][1] = circP->points0[i][1];

    circP->pointsR[circP->pointNo - i - 1][0] = circP->points0[i][0];
    circP->pointsR[circP->pointNo - i - 1][1] = circP->points0[i][1];
    circP->pointsR[circP->pointNo - i - 1][2] = 0.0;
  }

}

static CircData *
getCirc(int prec)
{
  int i;

  if (prec > MaxPrec) {
    if (MaxPrec < 0)
      Circles = malloc((prec + 1) * sizeof(*Circles));
    else
      Circles = realloc(Circles, (prec + 1) * sizeof(*Circles));

    for (i = MaxPrec + 1; i <= prec; i++)
      calcCirc(Circles + i, i);

    MaxPrec = prec;
  }

  return Circles + prec;
}

void
drawCircle2D(float x[2], float r, BOOL filled)
{
  CircData *circP;
  int i;

  /* Use full precision here because 2D circles normally end up
     larger than 3D circles. They're rarely used anyway, so it
     doesn't matter if we use high precision. */
  circP = getCirc(SgGetPrecision());

  for (i = 0; i < circP->pointNo; i++) {
    circP->points2D[i][0] = r * circP->points0[i][0] + x[0];
    circP->points2D[i][1] = r * circP->points0[i][1] + x[1];
  }

  if (filled)
    SgDrawPolygon2D(circP->points2D, circP->pointNo - 1);
  else
    SgDrawPolyline2D(circP->points2D, circP->pointNo);
}

void
SgDefDrawCircle2D(float x[2], float r)
{
  drawCircle2D(x, r, FALSE);
}

void
SgDefDrawDisc2D(float x[2], float r)
{
  drawCircle2D(x, r, TRUE);
}

static void
setTransf(float x[3], float v[3])
{
  float a, m[4][4];

  a = sqrtf(v[0] * v[0] + v[1] * v[1]);
  if (a > EpsDist) {
    m[0][0] = v[0] * v[2] / a;
    m[0][1] = v[1] * v[2] / a;
    m[0][2] = - a;
    m[0][3] = 0.0;
    m[1][0] = - v[1] / a;
    m[1][1] = v[0] / a;
    m[1][2] = 0.0;
    m[1][3] = 0.0;
    m[2][0] = v[0];
    m[2][1] = v[1];
    m[2][2] = v[2];
    m[2][3] = 0.0;
    m[3][0] = 0.0;
    m[3][1] = 0.0;
    m[3][2] = 0.0;
    m[3][3] = 1.0;
  }
  SgTranslate(x);
  if (a > EpsDist)
    SgMultMatrix(m);
}

static void
drawCircle(float x[3], float v[3], float r, BOOL filled)
{
  CircData *circP;
  float a;

  circP = getCirc(DefGetSpherePrecision(r));

  a = sqrtf(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
  v[0] /= a;
  v[1] /= a;
  v[2] /= a;
  SgPushMatrix();
  setTransf(x, v);
  SgScale(r, r, r);
  if (filled)
    SgDrawPolygon(circP->points0, circP->pointNo - 1, v);
  else
    SgDrawPolyline(circP->points0, circP->pointNo);
  SgPopMatrix();
}

void
SgDefDrawCircle(float x[3], float v[3], float r)
{
  drawCircle(x, v, r, FALSE);
}

void
SgDefDrawDisc(float x[3], float v[3], float r)
{
  drawCircle(x, v, r, TRUE);
}

void
SgDefDrawCylinder(float x0[3], float x1[3], float r,
    SgCylinderEnd end0, SgCylinderEnd end1)
{
  int prec;
  CircData *circP;
  float xm[4][3], xm2[4][2];
  float v[3], a, t[3], nv[3];
  int i;

  prec = DefGetSpherePrecision(r);
  circP = getCirc(prec);

  if (SgGetFeature(SG_FEATURE_FAST_SPHERE)) {
    if (SgMapCylinder(xm, NULL, NULL, x0, x1, r) > 0) {
      for (i = 0; i < 4; i++) {
	xm2[i][0] = xm[i][0];
	xm2[i][1] = xm[i][1];
      }
      SgDrawPolygon2D(xm2, 4);
    }
    return;
  }

  v[0] = x1[0] - x0[0];
  v[1] = x1[1] - x0[1];
  v[2] = x1[2] - x0[2];
  a = sqrtf(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
  v[0] /= a;
  v[1] /= a;
  v[2] /= a;
  SgPushMatrix();
  setTransf(x0, v);

  for (i = 0; i < circP->pointNo; i++)
    circP->points1[i][2] = a / r;

  SgScale(r, r, r);
  SgDrawStrip(circP->points1, circP->points0,
      circP->points0, NULL, circP->pointNo);

  if (end0 == SG_CYLINDER_CAP_FLAT) {
    nv[0] = 0.0;
    nv[1] = 0.0;
    nv[2] = -1.0;
    SgDrawPolygon(circP->pointsR, circP->pointNo - 1, nv);
  } else if (end0 == SG_CYLINDER_CAP_ROUND) {
    DefDrawHalfSphere(FALSE, prec);
  }

  if (end1 == SG_CYLINDER_CAP_FLAT) {
    nv[0] = 0.0;
    nv[1] = 0.0;
    nv[2] = 1.0;
    SgDrawPolygon(circP->points1, circP->pointNo - 1, nv);
  } else if (end1 == SG_CYLINDER_CAP_ROUND) {
    t[0] = 0.0;
    t[1] = 0.0;
    t[2] = a / r;
    SgTranslate(t);
    DefDrawHalfSphere(TRUE, prec);
  }

  SgPopMatrix();
}
